#!/usr/bin/env python
# coding: utf-8

# In[ ]:


"""
Provides a spark-based X-learner heterogeneous treatment effect estimator.
"""

from pyspark.sql.functions import monotonically_increasing_id
from pyspark.ml.feature import VectorAssembler
from pyspark.sql.functions import avg
from pyspark.sql.functions import lit
from pyspark.sql.functions import col
from pyspark.sql.functions import when
from pyspark.sql import SparkSession
from pyspark.sql.functions import udf
from pyspark.sql.types import FloatType
    
class SparkXLearner:
    """
    Spark-based X-learner heterogeneous treatment effect estimator.

    Assumptions
    ---------------
    This class assumes that the data is already stored in a distributed storage system (e.g., HDFS).
    This class also assumes that the treatment variable only contains 1s and 0s.
    """
    
    def __init__(self, learner="T"):
        self.treatments = []  # Multiple treatment effects can be estimated
        self.covariates = []
        self.outcome = None
        self.estimator_10 = None
        self.estimator_11 = None
        self.estimator_20 = None
        self.estimator_21 =  None
        self.propensity_estimator = None

    def fit(self, data, treatments, outcome, estimator_10, estimator_11, estimator_20, estimator_21, propensity_estimator):
        """
        Wrapper function to fit an ML-based counterfacual model.
        When multiple treatments are inputted, each treatment effect is estiamted individually.
        
        Parameters
        ----------
        data (2-D Spark dataframe): Base dataset containing features, treatment, iv, and outcome variables
        treatments (List): Names of the treatment variables             
        outcome (Str): Name of the outcome variable
        estimator_10 (mllib model obj): Arbitrary ML model of choice
        estimator_11 (mllib model obj): Arbitrary ML model of choice
        estimator_20 (mllib model obj): Arbitrary ML model of choice
        estimator_21 (mllib model obj): Arbitrary ML model of choice
        propensity_estimator (mllib model obj): Arbitrary ML model for propensity function
        
        Returns
        ------
        self
        """
        
        self.treatments = treatments
        self.outcome = outcome
        self.covariates = [var for var in data.columns if var not in treatments and var != outcome]
        self.estimator_10 = estimator_10
        self.estimator_11 = estimator_11
        self.estimator_20 = estimator_20
        self.estimator_21 =  estimator_21
        self.propensity_estimator = propensity_estimator
        self.__fit(data)
            
    def effects(self, X, treatment):
        """
        Function to get the estimated heterogeneous treatment effect from the fitted counterfactual model.
        
        The treatment effect is calculated by taking the difference between the predicted counterfactual outcomes.
        
        Parameters
        ----------
        X (2-D Spark dataframe): Feature data to estimate treatment effect of
        treatment (Str): Name of the treatment variable   
        
        returns
        -------
        cate: conditional average treatment effect
        ate: average treatment effect
        """
        
        assert treatment in self.treatments
        
        # Final prediction
        assembler = VectorAssembler(inputCols=self.covariates+[treatment], outputCol='features')
        X_assembled = assembler.transform(X)
        prediction_21 = estimator_21.transform(X_assembled.select('features')).withColumnRenamed("prediction", "prediction_21").select("prediction_21")
        prediction_20 = estimator_20.transform(X_assembled.select('features')).withColumnRenamed("prediction", "prediction_20").select("prediction_20")

        # Propensity function
        propensity_assembler = VectorAssembler(inputCols=self.covariates, outputCol='features')
        treatment_group_prop = propensity_assembler.transform(X)
        treatment_group_prop = treatment_group_prop.select(['features', treatment])
        treatment_prob = propensity_estimator.transform(treatment_group_prop).select("probability")
        firstelement = udf(lambda v:float(v[1]),FloatType())
        treatment_prob = treatment_prob.select(firstelement('probability')).withColumnRenamed("<lambda>(probability)", "probability")

        # Get cate
        X_w_pred = self.__mergeDfCol(X, prediction_21)
        X_w_pred = self.__mergeDfCol(X_w_pred, prediction_20)
        X_w_pred = self.__mergeDfCol(X_w_pred, treatment_prob)
        X_w_pred = X_w_pred.withColumn("probability", X_w_pred.probability.cast("float"))
        
        # should be + but - seems to produce the correct results...
        self.cate[treatment] = X_w_pred.select((X_w_pred.probability * X_w_pred.prediction_21) - ((lit(1) - X_w_pred.probability) * X_w_pred.prediction_20)).withColumnRenamed("((probability * prediction_21) + ((1 - probability) * prediction_20))", "cate")
        self.average_treatment_effects[treatment] = float(self.cate[treatment].groupby().avg().head()[0])
            
        return cate, ate

    
    def __fit(self, data):
        
        # TODO: result of X-learner is a little weird...prediction20 and prediction21 are very close. Maybe need to check implementation again
        
        for treatment in self.treatments:
            
            # Set up assembler
            assembler = VectorAssembler(inputCols=self.covariates+[treatment], outputCol='features')
            
            # First Stage
            # First estimator (treatment group)
            treatment_group = data.filter(treatment+" == 1")
            treatment_group_assembled = assembler.transform(treatment_group)
            treatment_group_assembled = treatment_group_assembled.select(['features', self.outcome])
            self.estimator_11 = self.estimator_11.fit(treatment_group_assembled)
            
            # Second estimator (control group)
            control_group = data.filter(treatment+" == 0")
            control_group_assembled = assembler.transform(control_group)
            control_group_assembled = control_group_assembled.select(['features', self.outcome])
            self.estimator_10 = self.estimator_10.fit(control_group_assembled)
            
            # Second stage
            # Get imputed counterfactuals
            counterfactual_control = estimator_11.transform(control_group_assembled.select('features')).withColumnRenamed("prediction", "prediction_10").select("prediction_10")
            counterfactual_treatment = estimator_10.transform(treatment_group_assembled.select('features')).withColumnRenamed("prediction", "prediction_11").select("prediction_11")
            
            # Get imputed treatment effect
            # TODO: df.select(self.outcome) doesn't seem to work here for some reason...let's fix the label column as "label" for now when inputting
            treatment_group_assembled_cf = self.__mergeDfCol(treatment_group_assembled, counterfactual_treatment)
            treatment_group_assembled_cate = treatment_group_assembled_cf.select(treatment_group_assembled_cf.outcome - treatment_group_assembled_cf.prediction_11).withColumnRenamed("(outcome - prediction_11)", "imputed_treatment")
            control_group_assembled_cf = self.__mergeDfCol(control_group_assembled, counterfactual_control)
            control_group_assembled_cate = control_group_assembled_cf.select(control_group_assembled_cf.outcome - control_group_assembled_cf.prediction_10).withColumnRenamed("(outcome - prediction_10)", "imputed_treatment")
            
            # Third stage
            treatment_group_third = self.__mergeDfCol(treatment_group_assembled, treatment_group_assembled_cate)
            control_group_third = self.__mergeDfCol(control_group_assembled, control_group_assembled_cate)
            self.estimator_21 = self.estimator_21.fit(treatment_group_third)
            self.estimator_20 = self.estimator_20.fit(control_group_third)
            
            # Fit propensity estimator
            assembler_propensity = VectorAssembler(inputCols=self.covariates, outputCol='features')
            treatment_group_prop = assembler_propensity.transform(data)
            treatment_group_prop = treatment_group_prop.select(['features', treatment])
            self.propensity_estimator = self.propensity_estimator.fit(treatment_group_prop)
            treatment_prob = propensity_estimator.transform(treatment_group_prop).select("probability")
            firstelement=udf(lambda v:float(v[1]),FloatType())
            treatment_prob = treatment_prob.select(firstelement('probability')).withColumnRenamed("<lambda>(probability)", "probability")
            

    def __mergeDfCol(self, df_1, df_2):
        """
        Function to merge two spark dataframes.
        
        Parameters
        ----------
        df_1 (2-D Spark dataframe): Spark dataframe to merge 
        df_2 (2-D Spark dataframe): Spark dataframe to merge
        
        Returns
        ------
        df_3 (2-D Spark dataframe): Spark dataframe merged by df1 and df2
        """
        
        df_1 = df_1.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_2 = df_2.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_3 = df_2.join(df1, "COL_MERGE_ID").drop("COL_MERGE_ID")
        return df_3




#!/usr/bin/env python
# coding: utf-8

# In[1]:


"""
Provides a spark-based two-stage least squares treatment effect estimator.
"""

from pyspark.sql.functions import monotonically_increasing_id
from pyspark.ml.feature import VectorAssembler
from pyspark.sql.functions import avg
from pyspark.sql.functions import lit
from pyspark.sql.functions import when
from pyspark.sql import SparkSession
from pyspark.ml.regression import LinearRegression

class SparkSieveTSLS:
    """
    Spark-based two-stage least squared heterogeneous treatment effect estimator.
    
    Assumptions
    ---------------
    This class assumes that the data is already stored in a distributed storage system (e.g., HDFS).
    This class also assumes that the treatment variable only contains 1s and 0s.
    """

    def __init__(self):
        self.treatments = []  # Multiple treatment effects can be estimated
        self.covariates = []
        self.outcome = ""
        self.iv = ""
        self.ols_1 = LinearRegression(featuresCol = 'features', labelCol=treatment)
        self.ols_2 = LinearRegression(featuresCol = 'features', labelCol=self.outcome)

    def fit(self, data, treatments, outcome, iv):
        """
        Wrapper function to fit first and second stage linear model to get adjusted treatment and a counterfacual model.
        When multiple treatments are inputted, each treatment effect is estiamted individually.
        
        Parameters
        ----------
        data (2-D Spark dataframe): Base dataset containing features, treatment, iv, and outcome variables
        treatments (List): Names of the treatment variables             
        outcome (Str): Name of the outcome variable
        iv (Str): Name of the instrument variable
              
        Returns
        ------
        self
        """
        
        self.treatments = treatments
        self.outcome = outcome
        self.iv = iv
        self.covariates = [var for var in data.columns if var not in treatments and var != iv and var != outcome]
        self.__fit(data)
    
    def effects(self, X, treatment):
        """
        Function to get the estimated heterogeneous treatment effect from the fitted counterfactual model.
        
        The treatment effect is calculated by taking the difference between the predicted counterfactual outcomes.
        
        Parameters
        ----------
        X (2-D Spark dataframe): Feature data to estimate treatment effect of
        treatment (Str): Name of the treatment variable   
        
        returns
        -------
        cate: conditional average treatment effect
        ate: average treatment effect
        """
        
        # Input treatment has to be fitted
        assert treatment in self.treatments
        
        # Predict adjusted treatment
        temp_1 = self.__mergeDfCol(X.select(treatment), X.select(self.iv))
        assembler = VectorAssembler(inputCols=[self.iv], outputCol='features')
        temp_1_assembled = assembler.transform(temp_1)
        temp_1_assembled = temp_1_assembled.select(['features', treatment])
        adjustedTreatment = self.ols_1 .transform(temp_1_assembled.select('features')).select("prediction")
        
        # Get predicted counterfactual for Y1 and Y0
        X_adjusted = self.__mergeDfCol(X, adjustedTreatment)
        counterfactual_treatment = X_adjusted.withColumn("prediction", lit(1))
        counterfactual_control =  X_adjusted.withColumn("prediction", lit(0))
        counterfactual_treatment_assembled = assembler.transform(counterfactual_treatment).select("features")
        counterfactual_control_assembled = assembler.transform(counterfactual_control).select("features")
        prediction_1 = self.ols_2.transform(counterfactual_treatment_assembled).withColumnRenamed("prediction", "prediction_1").select("prediction_1")
        prediction_0 = self.ols_2.transform(counterfactual_control_assembled).withColumnRenamed("prediction", "prediction_0").select("prediction_0")

        # Get cate and ate
        data_w_pred = self.__mergeDfCol(data, prediction_1)
        data_w_pred = self.__mergeDfCol(data_w_pred, prediction_0)
        cate = data_w_pred.select(data_w_pred.prediction_1 - data_w_pred.prediction_0).withColumnRenamed("(prediction_1 - prediction_0)", "cate")
        ate = float(cate.groupby().avg().head()[0])  # simply the average of cate
        return cate, ate
            
    def __fit(self, data):
        for treatment in self.treatments:
            
            # Fit first stage OLS
            trainSet_1 = self.__mergeDfCol(data.select(treatment), data.select(self.iv))
            assembler = VectorAssembler(inputCols=[self.iv], outputCol='features')
            trainSet_1_assembled = assembler.transform(trainSet_1)
            trainSet_1_assembled = trainSet_1_assembled.select(['features', treatment])
            self.ols_1 = self.ols_1.fit(trainSet_1_assembled)
            adjustedTreatment = self.ols_1 .transform(trainSet_1_assembled.select('features')).select("prediction")
            
            # Fit second stage OLS
            trainSet_2 = self.__mergeDfCol(data, adjustedTreatment)
            assembler = VectorAssembler(inputCols=self.covariates+["prediction"], outputCol='features')
            trainSet_2_assembled = assembler.transform(trainSet_2)
            trainSet_2_assembled = trainSet_2_assembled.select(['features', self.outcome])
            self.ols_2 = self.ols_2.fit(trainSet_2_assembled)


    def __mergeDfCol(self, df_1, df_2):
        """
        Function to merge two spark dataframes.
        
        Parameters
        ----------
        df_1 (2-D Spark dataframe): Spark dataframe to merge 
        df_2 (2-D Spark dataframe): Spark dataframe to merge
        
        Returns
        ------
        df_3 (2-D Spark dataframe): Spark dataframe merged by df1 and df2
        """
        
        df_1 = df_1.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_2 = df_2.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_3 = df_2.join(df1, "COL_MERGE_ID").drop("COL_MERGE_ID")
        return df_3




#!/usr/bin/env python
# coding: utf-8

# In[2]:


"""
Provides a spark-based S-learner heterogeneous treatment effect estimator.
"""

from pyspark.sql.functions import monotonically_increasing_id
from pyspark.ml.feature import VectorAssembler
from pyspark.sql.functions import avg
from pyspark.sql.functions import lit
from pyspark.sql.functions import col
from pyspark.sql.functions import when
from pyspark.sql import SparkSession
from pyspark.sql.functions import udf
from pyspark.sql.types import FloatType
    
class SparkSLearner:
    """
    Spark-based S-learner heterogeneous treatment effect estimator.
    
    This class currently only supports numeric outcomes.
    
    Assumptions
    ---------------
    This class assumes that the data is already stored in a distributed storage system (e.g., HDFS).
    This class also assumes that the treatment variable only contains 1s and 0s.
    """
    
    def __init__(self):
        self.treatments = []  # Multiple treatment effects can be estimated
        self.covariates = []
        self.outcome = ""
        self.estimator = None
        
    def fit(self, data, treatments, outcome, estimator):
        """
        Wrapper function to fit an ML-based counterfacual model.
        When multiple treatments are inputted, each treatment effect is estiamted individually.
        
        Parameters
        ----------
        data (2-D Spark dataframe): Base dataset containing features, treatment, iv, and outcome variables
        treatments (List): Names of the treatment variables             
        outcome (Str): Name of the outcome variable
        estimator (mllib model obj): Arbitrary ML estimator of choice
              
        Returns
        ------
        self
        """
        
        self.treatments = treatments
        self.outcome = outcome
        self.covariates = [var for var in data.columns if var not in treatments and var != outcome]
        self.estimator = estimator
        self.__fit(data)

    def effects(self, X, treatment):
        """
        Function to get the estimated heterogeneous treatment effect from the fitted counterfactual model.
        
        The treatment effect is calculated by taking the difference between the predicted counterfactual outcomes.
        
        Parameters
        ----------
        X (2-D Spark dataframe): Feature data to estimate treatment effect of
        treatment (Str): Name of the treatment variable   
        
        returns
        -------
        cate: conditional average treatment effect
        ate: average treatment effect
        """
        
        # Input treatment has to be fitted
        assert treatment in self.treatments
        
        # Get predictions for treatment and control group
        counterfactual_treatment = X.withColumn(treatment, lit(1))
        counterfactual_control =  X.withColumn(treatment, lit(0))
        assembler = VectorAssembler(inputCols=self.covariates+[treatment], outputCol='features')
        counterfactual_treatment_assembled = assembler.transform(counterfactual_treatment).select("features")
        counterfactual_control_assembled = assembler.transform(counterfactual_control).select("features")
        prediction_1 = self.estimator.transform(counterfactual_treatment_assembled).withColumnRenamed("prediction", "prediction_1").select("prediction_1")
        prediction_0 = self.estimator.transform(counterfactual_control_assembled).withColumnRenamed("prediction", "prediction_0").select("prediction_0")

        # Get cate
        X_w_pred = self.__mergeDfCol(X, prediction_1)
        X_w_pred = self.__mergeDfCol(X_w_pred, prediction_0)
        self.cate[treatment] = X_w_pred.select(X_w_pred.prediction_1 - X_w_pred.prediction_0).withColumnRenamed("(prediction_1 - prediction_0)", "cate")
        self.average_treatment_effects[treatment] = float(self.cate[treatment].groupby().avg().head()[0])
            
        return cate, ate

    def __fit(self, data):
        for treatment in self.treatments:
            
            # Single estimator
            assembler = VectorAssembler(inputCols=self.covariates+[treatment], outputCol='features')
            data_assembled = assembler.transform(data)
            data_assembled = data_assembled.select(['features', self.outcome])
            self.estimator = self.estimator.fit(data_assembled)
        
    def __mergeDfCol(self, df_1, df_2):
        """
        Function to merge two spark dataframes.
        
        Parameters
        ----------
        df_1 (2-D Spark dataframe): Spark dataframe to merge 
        df_2 (2-D Spark dataframe): Spark dataframe to merge
        
        Returns
        ------
        df_3 (2-D Spark dataframe): Spark dataframe merged by df1 and df2
        """
        
        df_1 = df_1.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_2 = df_2.withColumn("COL_MERGE_ID", monotonically_increasing_id())
        df_3 = df_2.join(df1, "COL_MERGE_ID").drop("COL_MERGE_ID")
        return df_3



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="reina",
    version="0.0.2",
    author="Qyu.ai Inc.",
    author_email="soumil@qyu.ai",
    description="A Causal Inference library for Big Data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Qyu-ai/Reina/",
    project_urls={
        "Bug Tracker": "https://github.com/Qyu-ai/Reina/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=['py4j==0.10.9.2', 'pyspark'],
    extras_require={
        'ml': ['numpy>=1.7'],
        'mllib': ['numpy>=1.7'],
        'sql': [
            'pandas>=0.23.2',
            'pyarrow>=1.0.0',
        ],
        'pandas_on_spark': [
            'pandas>=0.23.2',
            'pyarrow>=1.0.0',
            'numpy>=1.14',
        ],
    }
)
